#include "eopcc/las_io.hpp"
#include "eopcc/ply_io.hpp"


#include <sstream>

namespace eopcc {



template<> uint8_t ReadFromAscii<uint8_t>(std::ifstream& in_file) {
  int tval;
  in_file >> tval;
  return tval;
}

PLYReader::PLYReader(const std::string& filename) {
  ply_file_.open(filename, std::ifstream::binary);

  ParseHeader();
}

void PLYReader::ParseHeader() {
  std::string line;
  getline(ply_file_, line);

  if (line != "ply") {
    throw std::runtime_error("File is not PLY");
  }

  bool found_element = false;
  int property_number = 0;
  std::vector<std::string> property_names;
  std::vector<std::string> property_types;

  getline(ply_file_, line);
  while (line != "end_header") {
    std::stringstream sstream(line);
    std::string label, type, name;

    sstream >> label >> type >> name;

    if (label == "format") {
      if (type == "binary_little_endian") {
        is_binary_ = true;
      } else if (type == "binary_big_endian") {
        is_binary_ = true;
        throw std::runtime_error("Big endian not yet supported");
      } else if (type == "ascii") {
        is_binary_ = false;
      } else {
        throw std::runtime_error("Unrecognized format " + type);
      }
    } else if (label == "element") {
      if (type != "vertex") {
        throw std::runtime_error("Cannot load non-vertex elements into point cloud");
      }
      if (!found_element) {
        #ifdef WIN32
        num_points_remaining_ = _atoi64(name.c_str());
        #else
        num_points_remaining_ = std::atoll(name.c_str());
        #endif
        found_element = true;
      } else {
        throw std::runtime_error("Point cloud load only supports one element per PLY");
      }

    } else if (label == "property") {
      property_types.push_back(type);
      property_names.push_back(name);
    }

    getline(ply_file_, line);
  }

  num_properties_ = property_names.size();
  for (std::string& name : property_names) {
    property_attr_.push_back(StringToAttributeType(name));
  }

  for (std::string& type : property_types) {
    property_types_.push_back(StringToDataType(type));
  }

  SetupSeedCloud();
}

void PLYReader::SetupSeedCloud() {
  for (auto type : property_attr_) {
    if (type == AttributeType::Timestamp) {
      seed_cloud_.has_time = true;
    }

    if (type == AttributeType::Red || type == AttributeType::Green || type == AttributeType::Blue) {
      seed_cloud_.has_color = true;
    }
  }
}

void PLYReader::ReadPoint(PointCloud& pc, int index) {
  for (int i=0; i < num_properties_; ++i) {
    DataType type = property_types_[i];
    AttributeType attr = property_attr_[i];

    switch (type) {
      case DataType::uint8:
        ReadIntoCloud<uint8_t>(ply_file_, is_binary_, pc, index, attr);
        break;
      case DataType::uint16:
        ReadIntoCloud<uint16_t>(ply_file_, is_binary_, pc, index, attr);
        break;
      case DataType::int32:
        ReadIntoCloud<int32_t>(ply_file_, is_binary_, pc, index, attr);
        break;
      case DataType::int64:
        ReadIntoCloud<int64_t>(ply_file_, is_binary_, pc, index, attr);
        break;
      case DataType::float32:
        ReadIntoCloud<float>(ply_file_, is_binary_, pc, index, attr);
        break;
      case DataType::float64:
        ReadIntoCloud<double>(ply_file_, is_binary_, pc, index, attr);
        break;
    }
  }
}

PointCloud PLYReader::ReadSubCloud(int sub_cloud_size) {
  int num_to_read = std::min<size_t>(num_points_remaining_, sub_cloud_size);

  PointCloud pc = seed_cloud_;
  pc.resize(num_to_read);
  for (int i=0; i < num_to_read; ++i) {
    ReadPoint(pc, i);

    if (ply_file_.eof()) {
      std::cerr << "Early eof found, corrupt PLY?" << std::endl;
      break;
    }
  }
  num_points_remaining_ -= num_to_read;
  return pc;
}



void LoadFromPLY(const std::string& ply_file_name, PointCloud& point_cloud) {
  PLYReader reader(ply_file_name);
  point_cloud = reader.ReadSubCloud(reader.NumRemainingPoints());

}

template<typename T> void WriteBinary(std::ofstream& out, T val) {
  out.write(reinterpret_cast<char*>(&val), sizeof(T));
}
template<typename T> void WriteAscii(std::ofstream& out, T val) {
  out << val << " ";
}

template<> void WriteAscii<unsigned char>(std::ofstream& out, unsigned char val) {
  out << (int) val << " ";
}

template<typename T> void WriteAuto(std::ofstream& out, T val, bool is_binary) {
  if (is_binary) {
    WriteBinary(out, val);
  } else {
    WriteAscii(out, val);
  }
}


PLYWriter::PLYWriter(const std::string& filename, const PointCloud& seed_cloud, bool is_binary) {
  out_file_.open(filename, std::ofstream::binary);

  out_file_ << "ply\n";
  is_binary_ = is_binary;
  if (is_binary) {
    out_file_ << "format binary_little_endian 1.0\n";
  } else {
    out_file_ << "format ascii 1.0\n";
  }
  out_file_ << "element vertex ";
  header_count_pos_ = out_file_.tellp();
  out_file_ << "0\n";
  out_file_ << "comment padding for later rewrite of the vertex count\n";

  out_file_ << "property double x\n";
  out_file_ << "property double y\n";
  out_file_ << "property double z\n";
  out_file_ << "property float intensity\n";

  if (seed_cloud.has_time) {
    out_file_ << "property double timestamp\n";
    has_time_ = true;
  } else {
    has_time_ = false;
  }

  if (seed_cloud.has_color) {
    out_file_ << "property uchar red\n";
    out_file_ << "property uchar green\n";
    out_file_ << "property uchar blue\n";
    has_color_ = true;
  } else {
    has_color_ = false;
  }
  points_written_ = 0;
  out_file_ << "end_header\n";
  WriteSubCloud(seed_cloud);
}

PLYWriter::~PLYWriter() {
  out_file_.seekp(header_count_pos_);

  out_file_ << points_written_ << "\ncomment here... ";
}

void PLYWriter::WriteSubCloud(const PointCloud& pc) {
    for (int i=0; i < pc.size(); ++i) {
      WriteAuto<double>(out_file_, pc.loc[i].x, is_binary_);
      WriteAuto<double>(out_file_, pc.loc[i].y, is_binary_);
      WriteAuto<double>(out_file_, pc.loc[i].z, is_binary_);

      WriteAuto<float>(out_file_, pc.intensity[i], is_binary_);

      if (has_time_) {
          WriteAuto<double>(out_file_, pc.gps_time[i], is_binary_);
      }

      if (has_color_) {
          WriteAuto<unsigned char>(out_file_, pc.color[i].red * 255.0, is_binary_);
          WriteAuto<unsigned char>(out_file_, pc.color[i].green * 255.0, is_binary_);
          WriteAuto<unsigned char>(out_file_, pc.color[i].blue * 255.0, is_binary_);
      }

      if (!is_binary_) {
        out_file_ << "\n";
      }
    }

    points_written_ += pc.size();
}



void WriteToPLY(const std::string& filename, const PointCloud& point_cloud, bool is_binary) {
  PLYWriter ply_writer(filename, point_cloud, is_binary);

}

} // end namespace eopcc
